package de.hhn.fhir;

import java.time.LocalDate;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Random;

import org.hl7.fhir.dstu3.model.*;
import org.hl7.fhir.dstu3.model.Enumerations.AdministrativeGender;

import ca.uhn.fhir.context.FhirContext;
import ca.uhn.fhir.rest.api.MethodOutcome;
import ca.uhn.fhir.rest.client.api.IGenericClient;

public class FhirTutorialDortmund2017 {
	private static IGenericClient client;
	private static FhirContext ctx;
	private static Patient patientAusAufgabe3;

	public static void main(String[] args) throws InterruptedException {
		erzeugeFhirContext();
		patientenErzeugen();
		suchePatientenMitNamen();
		updatePatient();
		komplexePatientenSuche();
		glucoseObservationErzeugen();
	}

	private static void erzeugeFhirContext() {
		ctx = FhirContext.forDstu3();
		String serverBase = "https://fhir.molit.eu/baseDstu3";
		client = ctx.newRestfulGenericClient(serverBase);
	}

	private static void patientenErzeugen() {

		String identSystem = "http://dit2017.de";
		String identValue = "RickSanchez";

		Patient patient = new Patient();
		patient.addIdentifier().setSystem(identSystem).setValue(identValue);
		patient.addName().addGiven("Rick").setFamily("Sanchez");
		patient.addAddress().setCity("Dortmund").addLine("Musterstraße 123");
		patient.setGender(AdministrativeGender.MALE);

		Date date = Date.from(LocalDate.of(1947, 1, 15).atStartOfDay(ZoneId.systemDefault()).toInstant());
		patient.setBirthDate(date);

		// Patient wird hier nur erzeugt falls er noch nicht existiert.
		MethodOutcome outcome = client.create().resource(patient).conditional()
				.where(Patient.IDENTIFIER.exactly().systemAndIdentifier(identSystem, identValue)).execute();

		String id = outcome.getId().getValue();
		String stringXmlPat = ctx.newXmlParser().setPrettyPrint(true).encodeResourceToString(patient);
		System.out.println(stringXmlPat);
		System.out.println("Patient created? " + outcome.getCreated() + " ID: " + id);
	}

	private static void suchePatientenMitNamen() {

		Bundle results = client.search().forResource(Patient.class).where(Patient.NAME.matches().value("Sanchez"))
				.returnBundle(Bundle.class).execute();

		System.out.println("Found " + results.getEntry().size());
		String bundleString = ctx.newXmlParser().setPrettyPrint(true).encodeResourceToString(results);
		System.out.println(bundleString);

		Patient patient = (Patient) results.getEntryFirstRep().getResource();
		patientAusAufgabe3 = patient;

		String vornamen = patient.getNameFirstRep().getGivenAsSingleString();
		System.out.println("Vornamen & Nachname: " + vornamen + " " + patient.getNameFirstRep().getFamily());
	}

	private static void updatePatient() {

		List<Address> addresses = new ArrayList<>();
		Address address = new Address();
		address.setCity("Bochum");
		addresses.add(address);
		patientAusAufgabe3.setAddress(addresses);

		// Updated den Patienten mit der gleichen ID (patientAusAufgabe3 hat bereits
		// eine ID)
		MethodOutcome outcome = client.update().resource(patientAusAufgabe3).prettyPrint().encodedJson().execute();

		String id = outcome.getId().getValue();
		System.out.println("Got ID: " + id);

		// Updated den Patienten NICHT anhand der ID, hier wird der Patient mit gleichem
		// Identifier geupdated.
		MethodOutcome outcome2 = client.update().resource(patientAusAufgabe3).conditional()
				.where(Patient.IDENTIFIER.exactly().systemAndIdentifier(
						patientAusAufgabe3.getIdentifierFirstRep().getSystem(),
						patientAusAufgabe3.getIdentifierFirstRep().getValue()))
				.execute();
		System.out.println(outcome2.getId().getValue());
	}

	private static void komplexePatientenSuche() {

		Bundle response = client.search().forResource(Patient.class)
				.where(Patient.ADDRESS.matches().values("Bochum", "Dortmund"))
				.and(Patient.NAME.matches().values("Sanchez", "Morty", "oderDritterName"))
				.and(Patient.DECEASED.isMissing(false)).returnBundle(Bundle.class).execute();

		String bundleString = ctx.newXmlParser().setPrettyPrint(true).encodeResourceToString(response);
		System.out.println(bundleString);
	}

	private static void glucoseObservationErzeugen() throws InterruptedException {

		Patient patient = condCreateGlucPatient();

		for (int i = 0; i < 10; i++) {
			Observation observation = new Observation();
			observation.setSubject(new Reference()
					.setReference(patient.getIdElement().getResourceType() + "/" + patient.getIdElement().getIdPart())
					.setIdentifier(patient.getIdentifierFirstRep()).setDisplay(patient.getNameFirstRep().getText()));

			observation.setEffective(new DateTimeType(new Date()));
			observation.setStatus(Observation.ObservationStatus.FINAL);

			observation.setValue(new Quantity().setValue(new Random().nextInt(150) + 50)
					.setSystem("http://unitsofmeasure.org").setUnit("mg/dl").setCode("mg/dl"));

			System.out.println(ctx.newXmlParser().setPrettyPrint(true).encodeResourceToString(observation));
			client.create().resource(observation).execute();
			Thread.sleep(500);
		}
	}

	private static Patient condCreateGlucPatient() {

		Patient patient = new Patient();
		patient.addIdentifier().setSystem("http://dit2017.de").setValue("glucPatient");
		patient.addName().addGiven("Morty").setFamily("Smith");
		patient.addAddress().setCity("Dortmund").addLine("Musterstraße 123");

		Date date = Date.from(LocalDate.of(2014, 4, 24).atStartOfDay(ZoneId.systemDefault()).toInstant());
		patient.setBirthDate(date);

		MethodOutcome outcome = client.create().resource(patient).conditional()
				.where(Patient.IDENTIFIER.exactly().systemAndIdentifier(patient.getIdentifierFirstRep().getSystem(),
						patient.getIdentifierFirstRep().getValue()))
				.execute();

		String id = outcome.getId().getValue();
		System.out.println("GlucPatient created? " + outcome.getCreated() + " ID of Patient: " + id);
		patient.setId(id);
		return patient;
	}
}
